#!/bin/bash

# Copyright (c) 2015-2021, NVIDIA CORPORATION.  All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#  * Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#  * Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#  * Neither the name of NVIDIA CORPORATION nor the names of its
#    contributors may be used to endorse or promote products derived
#    from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
# EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
# OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#
# odmfuse.sh: Fuse the target board.
#           odmfuse performs the best in L4T fuse environment.
#
# Usage: Place the board in recovery mode and run:
#
#       ./odmfuse.sh -i <TegraID> [options] TargetBoard
#
#       for more detail enter 'odmfuse.sh -h'
#
# Examples for Jetson Xavier:
#   1. Secure fuse with PKC, SBK and KEK2:
#      ./odmfuse.sh -i 0x19 -k <Key file> -S <SBK file> --KEK2 <KEK2 file> jetson-xavier
#   2. Secure fuse with PKC and KEK2:
#      ./odmfuse.sh -i 0x19 -k <Key file> --KEK2 <KEK2 file> jetson-xavier
#   3. Burn SecurityMode so that no manufacturing fuses can be burned anymore:
#      ./odmfuse.sh -i 0x19 -p -k <Key file> -S <SBK file> jetson-xavier   # If the board is secured by PKC and SBK
#      ./odmfuse.sh -i 0x19 -p -k <Key file> jetson-xavier                 # If the board is secured by PKC only
#
# Examples for Jetson TX2:
#   1. Secure fuse with PKC, SBK and KEK2:
#      ./odmfuse.sh -i 0x18 -k <Key file> -S <SBK file> --KEK2 <KEK2 file> jetson-tx2
#   2. Secure fuse with PKC and KEK2:
#      ./odmfuse.sh -i 0x18 -k <Key file> --KEK2 <KEK2 file> jetson-tx2
#   3. Burn SecurityMode so that no manufacturing fuses can be burned anymore:
#      ./odmfuse.sh -i 0x18 -p -k <Key file> -S <SBK file> jetson-tx2   # If the board is secured by PKC and SBK
#      ./odmfuse.sh -i 0x18 -p -k <Key file> jetson-tx2                 # If the board is secured by PKC only
#
# Examples for Jetson TX1:
#   1. Secure fuse with PKC and encryption service preparation:
#      ./odmfuse.sh -i 0x21 -c PKC -p -k <KeyFile> -D <DK file> -S <SBK file>
#   2. Secure fuse with PKC:
#      ./odmfuse.sh -i 0x21 -c PKC -p -k <KeyFile>
#   3. Secure fuse without any boot authentication:
#      ./odmfuse.sh -i 0x21 -c NS -p
#
# Examples for Jetson TK1:
#   1. Secure fuse with PKC and encryption service preparation:
#      ./odmfuse.sh -i 0x40 -c PKC -p -k <KeyFile> -D <DK file> -S <SBK file>
#   2. Secure fuse with PKC:
#      ./odmfuse.sh -i 0x40 -c PKC -p -k <KeyFile>
#   3. Secure fuse without any boot authentication:
#      ./odmfuse.sh -i 0x40 -c NS -p
#

validateMaster ()
{
	if [[ "$1" =~ ^[^@]{1,}@[^@]{1,}$ ]]; then
		return 0;
	fi;
	echo "Error: dbmaster is not in <user>@<db server> format.";
	exit 1;
}

#
# Regardless of TegraID, all HEX inputs(DK, KEK, SBK, HASH) should be in
# single Big Endian format. This routine should not only check the
# input format but also convert to LE format.
#
chkhash ()
{
	local i;
	local keyname=$1;
	local keylen=$2;
	local keystr="${!keyname}";
	local le="";
	local resid=0;

	# 1. Check single HEX format.
	if [[ ${keystr} =~ ^0[xX] ]]; then
		keystr=${keystr:2};
	fi;
	keylen=$(( keylen * 2 ));
	if [ ${keylen} -ne ${#keystr} ]; then
		echo "Error: The length of ${keystr} = ${#keystr} != ${keylen}";
		exit 1;
	fi;
	resid=$((keylen % 8));
	if [ $resid -ne 0 ]; then
		echo "Error: ${keyname} length is not modulo 32bit";
		exit 1;
	fi;

	if [[ "${keystr}" = *[^[:alnum:]]* ]]; then
		echo "Error: ${keyname} has non-alphanumeric value";
		exit 1;
	fi;

	if [[ "${keystr}" =~ [g-zG-Z] ]]; then
		echo "Error: ${keyname} is not in HEX format.";
		exit 1;
	fi;

	# 2. Convert to Little Endian.
	# 3. Split to multiple 32 HEX string if necessary.
	if [ "${tid}" = "0x40" ]; then
		local tmpstr="";
		local n=$(( keylen + 7 ));
		n=$(( n / 8 ));
		local lastn=$((n - 1));
		for (( i=0; i<n; i++ )); do
			local s=$((i * 8));
			local tmple="${keystr:$s}";
			tmpstr+="0x${tmple:6:2}${tmple:4:2}";
			tmpstr+="${tmple:2:2}${tmple:0:2}";
			if [ $i -lt $lastn ]; then
				tmpstr+=" ";
			fi;
			keylen=$((keylen - 8));
		done;
		eval "${keyname}=\"${tmpstr}\"";
	elif [ "${tid}" != "0x18" -a "${tid}" != "0x19" ]; then
		i=${#keystr}
		while [ $i -gt 0 ]; do
			i=$[$i-2]
			le+="${keystr:$i:2}"
		done;
		eval "${keyname}=0x\"${le}\"";
	fi;
}

#
# Convert Little Endian hashes to single Big Endian HEX.
#
convhash ()
{
	local le="";
	local be="";
	local keyname=$1;
	local keylen=$2;
	local keystr="${!keyname}";

	# 1. Consolidate multiple hash tokens into single HEX format.
	if [ "${tid}" = "0x40" ]; then
		local keyarr="${keyname}arr";
		local n=$(( keylen * 2 ));
		n=$(( n + 7 ));
		n=$(( n / 8 ));
		OIFS=${IFS};
		IFS=' ';
		keyarr=(${keystr});
		IFS=${OIFS};
		if (( ${n}!=${#keyarr[@]} )); then
			echo "*** Error: bad ${keyname} length."
			exit 1;
		fi;

		for (( i=0; i<${n}; i++ )); do
			if ! [[ ${keyarr[$i]} =~ ^0x[0-9,A-F,a-f]{8} ]]; then
				echo "*** Error: bad ${keyname} element.";
				exit 1;
			fi;
			le+="${keyarr[$i]:8:2}${keyarr[$i]:6:2}";
			le+="${keyarr[$i]:4:2}${keyarr[$i]:2:2}";
		done;
		be=${le};
	else
		if [[ ${keystr} =~ ^0[xX] ]]; then
			le=${keystr:2};
		fi;
		local i=${#le}
		while [ $i -gt 0 ]; do
			i=$[$i-2]
			be+="${le:$i:2}"
		done;
	fi;

	# 2. Set the global variable with new value.
	eval "${keyname}=0x\"${be}\"";
}

factory_overlay_gen ()
{
	local fusecmd="$1";
	local fuseconf="$2";
	local tmpdir="";
	local workdir="";
	local cmdline="";
	local opts;
	local opt;
	local optnum;
	local i;
	local j;
	local files;
	local file;
	local filenum;
	local folder;

	pushd "${BL_DIR}" >& /dev/null;
	tmpdir="$(mktemp -d factory-overlay-tmpdir-XXXX)";
	if [ ! -d "${tmpdir}" ]; then
		echo "Error: Create temporary directory failed.";
		popd > /dev/null 2>&1;
		exit 1;
	fi;
	workdir="${tmpdir}/bootloader";
	mkdir "${workdir}";

	echo "*** Start preparing fuse configuration ... ";
	local fusecmdfile="fusecmd.sh"
	echo "#!/bin/bash" >  "${fusecmdfile}";
	echo "eval '${fusecmd}'" >> "${fusecmdfile}";
	chmod +x "${fusecmdfile}";
	mv "${fusecmdfile}" "${workdir}";

	cp -f "${fuseconf}" "${workdir}";
	rm -f *.raw;

	if [ "${tid}" = "0x40" ]; then
		cp nvflash "${workdir}";
		cp fastboot.bin "${workdir}";
	elif [ "${tid}" = "0x21" ]; then
		cp nvtboot_recovery.bin "${workdir}";
		cp tegraflash_internal.py "${workdir}";
		cp tegraflash.py "${workdir}";
		cp tegraparser "${workdir}";
		cp tegrarcm "${workdir}";
		cp tegrasign "${workdir}";
		# tegraflash_internal.py imports tegrasign_v3 unconditionally
		# so we have to package tegrasign_v3 files
		cp tegrasign_v3*.py "${workdir}";
	elif [ "${tid}" = "0x18" -o "${tid}" = "0x19" ]; then
		cp tegrabct_v2 "${workdir}";
		cp tegradevflash_v2 "${workdir}";
		cp tegraflash_internal.py "${workdir}";
		cp tegrahost_v2 "${workdir}";
		cp tegraparser_v2 "${workdir}";
		cp tegrarcm_v2 "${workdir}";
		cp tegrasign_v3*.py "${workdir}";
		cp tegraopenssl "${workdir}";
		if [ "${tid}" = "0x19" ]; then
			cp sw_memcfg_overlay.pl "${workdir}";
		fi;

		# Parsing the command line of tegraflash.py, to get all files that tegraflash.py and
		# tegraflash_internal.py needs so copy them to the working directory.
		cmdline=`echo "${fusecmd}" | sed -e s/\;/\ /g -e s/\"//g`;
		cmdline=`echo "${cmdline}" | sed -e s/\ \ */\ /g`;
		opts=($cmdline);
		optnum=${#opts[@]};
		for (( i=0; i < ${optnum}; )); do
			opt="${opts[$i]}";
			opt=`echo "${opt}" | sed -e s/\,/\ /g`;
			files=($opt);
			filenum=${#files[@]};
			for (( j=0; j < ${filenum}; )); do
				file="${files[$j]}";
				if [ -f "${file}" ]; then
					folder=$(dirname "${file}");
					if [ "${folder}" != "." ]; then
						mkdir -p "${workdir}/${folder}";
					fi;
					cp "${file}" "${workdir}/${folder}";
				fi;
				j=$((j+1));
			done;
			i=$((i+1));
		done;

		# If the flash.xml appears in the cmdline, that means we need not only files listed
		# in the cmdline, but also files in flash.xml. So parsing the flash.xml here.
		if [ -f "${workdir}/flash.xml" ]; then
			sed -i $'s/\t/    /g' "${workdir}/flash.xml";
			files=$(sed -n /^\ *\<filename/p "${workdir}"/flash.xml|sed -e s/^\ *//g -e s/\ \ */\ /g|cut -d" " -f 2);
			# Note: using "${files}" in the for loop is wrong because the double quotation marks
			# make the ${files} a single string with carriage returns
			for file in ${files}; do
				if [ "${file}" = "bmp.blob" ] || [ "${file}" = "system.img" ] ||
				   [ "${file}" = "slot_metadata.bin" ]; then
					continue;
				fi;
				if [ ! -f "${workdir}/${file}" ]; then
					cp "${file}" "${workdir}";
				fi;
			done;
		fi;
	else
		echo "Error: Not supported yet.";
		popd > /dev/null 2>&1;
		rm -rf "${tmpdir}";
		exit 1;
	fi;

	pushd "${tmpdir}" >& /dev/null;
	tar cjf fuseblob.tbz2 bootloader;
	mv fuseblob.tbz2 "${LDK_DIR}";
	popd > /dev/null 2>&1;

	rm -rf "${tmpdir}";
	popd > /dev/null 2>&1;
	echo "*** done.";
}

usage_t18x_t19x ()
{
	cat << EOF
Usage:
  ./odmfuse.sh -i <TegraID> [options] TargetBoard

  Where options are,
    -c <CryptoType> ---------- Set the crypto type(obsolete). Please use "--auth" instead.
    -d <0xXXXX> -------------- sets sec_boot_dev_cfg=0xXXXX&0x3fff.
    -i <TegraID> ------------- tegra ID: 0x18-TX2, 0x19-Xavier
    -j ----------------------- Keep jtag enabled (obsolete). Jtag by default is enabled.
                               Jtag can be disabled by using "--disable-jtag" option.
                               Jtag can't be re-enabled once the jtag disable fuse bit is burned.
    -k <KeyFile> ------------- 2048 bit RSA private KEY file. (.pem file)
    -l <0xXXX> --------------- sets odm_lock=0xX. (4 bits)
    -p ----------------------- sets production mode.
    -r <0xXX> ---------------- sets sw_reserved=0xXX.
    -S <SBK file> ------------ 128bit Secure Boot Key file in HEX format.
    --auth ------------------- Set the current authentication type of the board. Possible values:
                               NS -- No authentication, PKC - PKC is enabled, SBKPKC - SBK and PKC are enabled.
                               This option is only needed in offline mode, namely option "--noburn" is set.
    --noburn ----------------- Prepare fuse blob without actual burning.
    --test ------------------- No fuses will be really burned, for test purpose.
    --force ------------------ For fuses that have been burned, force to burn it again.
    --disable-jtag ----------- Burn the jtag-disable fuse. You can't re-enable it after it is burned.
    --KEK0 <Key file> -------- 128bit Key Encryption Key file in HEX format.
    --KEK1 <Key file> -------- 128bit Key Encryption Key file in HEX format.
    --KEK2 <Key file> -------- 128bit Key Encryption Key file in HEX format.
    --KEK256 <Key file> ------ 256bit Key Encryption Key file in HEX format.
    --odm_reserved[0-7] ------ sets 32bit ReservedOdm[0-7]. (Input=0xXXXXXXXX)
    --debug_authentication --- Set arm_debug_authentication=0xXX&0x1f. (5 bits)
    --odm_id ----------------- Set odm_id=0xXXXXXXXXXXXXXXXX. (64 bits)
                               High 32 bits: odm_id[0]. Low 32 bits: odm_id[1].
EOF
}

usage_t21x_t12x ()
{
	cat << EOF
Usage:
  ./odmfuse.sh -c <CryptoType> -i <TegraID> -k <KeyFile> [options]

  Where options are,
    -c <CryptoType> ------ NS -- No Crypto, PKC - Public Key Crypto.
    -d <0xXXXX> ---------- sets sec_boot_dev_cfg=0xXXXX&0x3fff.
    -i <TegraID> --------- tegra ID: 0x40-TK1, 0x21-TX1
    -j ------------------- Keep jtag enabled (obsolete). Jtag by default is enabled.
                           Jtag can be disabled by using "--disable-jtag" option.
                           Jtag can't be re-enabled once the jtag disable fuse bit is burned.
    -k <KeyFile> --------- 2048 bit RSA private KEY file. (.pem file)
    -l <0xX> ------------- sets odm_lock=0xX. (4 bits)
    -o <8-0xXXXXXXXX> ---- sets odm_reserved=<8-0xXXXXXXXX>
                           1 256bit Big Endian value.
    -p ------------------- sets production mode.
    -r <0xXX> ------------ sets sw_reserved=0xXX.
    -D <DK file> --------- 32bit Device Key file in HEX format (TK1 & TX1 only).
    -S <SBK file> -------- 128bit Secure Boot Key file in HEX format.
    --noburn ------------- Prepare fuse blob without actual burning.
    --test --------------- No fuses will be really burned, for test purpose.
    --disable-jtag ------- Burn the jtag-disable fuse. You can't re-enable it after it is burned.
EOF
}

usage_keyviadb ()
{
	cat << EOF
Usage: ./odmfuse.sh [options] -c <CryptoType> -i <Tegra ID> <master> <SN> [item]
  Where,
    -c <CryptoType> ------ NS -- No Crypto, PKC - Public Key Crypto.
    -d <0xXXXX> ---------- sets sec_boot_dev_cfg=0xXXXX&0x3fff.
    -i <TegraID> --------- tegra ID: 0x40-TK1, 0x21-TX1, 0x18-TX2, 0x19-Xavier
    -j ------------------- Keep jtag enabled (obsolete). Jtag by default is enabled.
                           Jtag can be disabled by using "--disable-jtag" option.
                           Jtag can't be re-enabled once the jtag disable fuse bit is burned.
    -l <0xX> ------------- sets odm_lock=0xX.
    -o <8-0xXXXXXXXX> ---- sets odm_reserved=<8-0xXXXXXXXX>
                           8 32bit values MUST be quoted.
    -p ------------------- sets production mode.
    -r <0xXX> ------------ sets sw_reserved=0xXX.
    -s <sku> ------------- Board SKU.
    --test --------------- No fuses will be really burned, for test purpose.
    --disable-jtag ------- Burn the jtag-disable fuse. You can't re-enable it after it is burned.

    master --------------- PKC DB master in <user>@<server> format.
    SN ------------------- Target board's Serial Number n arbitrary format.
    item ----------------- one DB column to be fetched:
                           id,sku,type,pub,hash,fusecnt,signcnt,ctime,atime
EOF
}

usage ()
{
	if [ "${KEYVIADB}" = "" ]; then
		if [ "${tid}" = "0x18" ] || [ "${tid}" = "0x19" ]; then
			usage_t18x_t19x;
		elif [ "${tid}" = "0x21" ] || [ "${tid}" = "0x40" ]; then
			usage_t21x_t12x;
		else
			echo "For Jetson TX2, Jetson Xavier, Jetson Xavier NX platforms:";
			usage_t18x_t19x;
			echo;
			echo "For Jetson TX1, Jetson TK1, Jetson nano platforms:";
			usage_t21x_t12x;
		fi;
	else
		usage_keyviadb;
	fi;
	exit 1;
}

check_ctype ()
{
	if [ "${tid}" = "0x40" ] || [ "${tid}" = "0x21" ]; then
		if [ "${Ctype}" = "" ]; then
			echo "*** Error: Crypto type is missing.";
			exit 1;
		fi;
		if [ "${Ctype}" != "NS" ] && [ "${Ctype}" != "PKC" ]; then
			echo "*** Error: unknown crypto type. (valid type = \"NS\" or \"PKC\")";
			exit 1;
		fi;
	fi;
	if [ "${tid}" = "0x18" ] || [ "${tid}" = "0x19" ]; then
		if [ "${running_mode}" = "${MODE_OFFLINE}" ]; then
			if [ "${Ctype}" = "" ]; then
				echo "*** Error: --auth is missing.";
				exit 1;
			fi;
			if [ "${Ctype}" != "NS" ] && [ "${Ctype}" != "PKC" ] && [ "${Ctype}" != "SBKPKC" ]; then
				echo "*** Error: illegal --auth type. (valid types = \"NS\" or \"PKC\" or \"SBKPKC\")";
				exit 1;
			fi;
		fi;
	fi;
}

get_running_mode ()
{
	if [ "${tid}" = "0x40" ] || [ "${tid}" = "0x21" ]; then
		if [ ${noburn} -eq 1 ]; then
			running_mode="${MODE_OFFLINE}";
		else
			running_mode="${MODE_ONLINE}";
		fi;
	fi;
	if [ "${tid}" = "0x18" ] || [ "${tid}" = "0x19" ]; then
		if [ ${noburn} -eq 0 ]; then
			running_mode="${MODE_ONLINE}";
		else
			if [ "${tid}" = "0x18" ]; then
				if [ "${FAB}" = "" ] || [ "${BOARDID}" = "" ]; then
					echo
					echo "Odmfuse requires variable FAB and BOARDID in order to run in the offline mode."
					echo "Otherwise odmfuse needs to access on board EEPROM. Make sure the board is in recovery mode."
					echo
					running_mode="${MODE_HYBRID}";
				else
					running_mode="${MODE_OFFLINE}";
				fi;
			fi;
			if [ "${tid}" = "0x19" ]; then
				if [ "${FAB}" = "" ] || [ "${BOARDID}" = "" ] || \
				   [ "${BOARDSKU}" = "" ] || [ "${BOARDREV}" = "" ]; then
					echo
					echo "Odmfuse requires variable FAB, BOARDID, BOARDSKU and BOARDREV in order to run in the offline mode."
					echo "Otherwise odmfuse needs to access on board EEPROM. Make sure the board is in recovery mode."
					echo
					running_mode="${MODE_HYBRID}";
				else
					if [ "${CHIPREV}" = "" ]; then
						echo "CHIPREV is missing, will use the default value(2).";
						CHIPREV="2";
					fi;
					running_mode="${MODE_OFFLINE}";
				fi;
			fi;
		fi;
	fi;
}

sanitize_inputs ()
{
	if [ "${tid}" = "" ]; then
		echo "*** Error: Tegra ID is missing.";
		usage;
	fi;
	if [ "${tid}" != "0x40" -a "${tid}" != "0x21" -a "${tid}" != "0x18" -a \
		"${tid}" != "0x19" ]; then
		echo "*** Error: Unsupported Tegra ID ${tid}.";
		exit 1;
	fi;

	if [ "${tid}" = "0x18" ] || [ "${tid}" = "0x19" ]; then
		if [ "${__ctype}" != "" ]; then
			echo;
			echo "The option -c is obsolete now. Please use \"--auth\" instead.";
			echo;
		fi;
	else
		Ctype="${__ctype}";
	fi;
	get_running_mode;
	check_ctype;
	if [ "${tid}" = "0x18" ] || [ "${tid}" = "0x19" ]; then
		if [ "${running_mode}" = "${MODE_OFFLINE}" ]; then
			check_sbk_pkc "${Ctype}" "${KEYFILE}" "${SBKFILE}";
		fi;
	fi;

	if [ "${DKFILE}" != "" ]; then
		if [ "${tid}" != "0x40" -a "${tid}" != "0x21" ]; then
			echo "Error: Device Key is not supported on T186 and T194 platforms.";
			exit 1;
		fi;
		if [ ! -f "${DKFILE}" ]; then
			echo "*** Error: ${DKFILE} doesn't exits.";
			exit 1;
		fi;
		DKFILE=`readlink -f "${DKFILE}"`;
		if [ "${SBKFILE}" = "" ]; then
			echo "*** Error: SBK is missing.";
			exit 1;
		fi;
	fi;

	if [ "${HASHFILE}" != "" ]; then
		if [ ! -f "${HASHFILE}" ]; then
			echo "*** Error: ${HASHFILE} doesn't exits.";
			exit 1;
		fi;
		HASHFILE=`readlink -f "${HASHFILE}"`;
		if [ "${KEYFILE}" != "" ]; then
			echo "*** Error: KEY and HASH are mutually exclusive.";
			exit 1;
		fi;
	fi;

	if [ "${KEK0FILE}" != "" ]; then
		if [ "${tid}" != "0x18" -a "${tid}" != "0x19" ]; then
			echo "Error: Key Encryption Key is supported only for TX2";
			exit 1;
		fi;
		if [ ! -f "${KEK0FILE}" ]; then
			echo "*** Error: ${KEK0FILE} doesn't exits.";
			exit 1;
		fi;
		KEK0FILE=`readlink -f "${KEK0FILE}"`;
	fi;

	if [ "${KEK1FILE}" != "" ]; then
		if [ "${tid}" != "0x18" -a "${tid}" != "0x19" ]; then
			echo "Error: Key Encryption Key is supported only for TX2";
			exit 1;
		fi;
		if [ ! -f "${KEK1FILE}" ]; then
			echo "*** Error: ${KEK1FILE} doesn't exits.";
			exit 1;
		fi;
		KEK1FILE=`readlink -f "${KEK1FILE}"`;
	fi;

	if [ "${KEK2FILE}" != "" ]; then
		if [ "${tid}" != "0x18" -a "${tid}" != "0x19" ]; then
			echo "Error: Key Encryption Key is supported only for TX2";
			exit 1;
		fi;
		if [ ! -f "${KEK2FILE}" ]; then
			echo "*** Error: ${KEK2FILE} doesn't exits.";
			exit 1;
		fi;
		KEK2FILE=`readlink -f "${KEK2FILE}"`;
	fi;

	if [ "${KEK256FILE}" != "" ]; then
		if [ "${tid}" != "0x18" -a "${tid}" != "0x19" ]; then
			echo "Error: Key Encryption Key is supported only for TX2";
			exit 1;
		fi;
		if [ ! -f "${KEK256FILE}" ]; then
			echo "*** Error: ${KEK256FILE} doesn't exits.";
			exit 1;
		fi;
		KEK256FILE=`readlink -f "${KEK256FILE}"`;
	fi;

	if [ "${KEYFILE}" != "" ]; then
		if [ ! -f "${KEYFILE}" ]; then
			echo "*** Error: ${KEYFILE} doesn't exits.";
			exit 1;
		fi;
		KEYFILE=`readlink -f "${KEYFILE}"`;
		if [ "${HASHFILE}" != "" ]; then
			echo "*** Error: HASH and KEY are mutually exclusive.";
			exit 1;
		fi;
	fi;

	if [ "${SBKFILE}" != "" ]; then
		if [ ! -f "${SBKFILE}" ]; then
			echo "*** Error: ${SBKFILE} doesn't exits.";
			exit 1;
		fi;
		SBKFILE=`readlink -f "${SBKFILE}"`;
		if [ "${DKFILE}" = "" ]; then
			if [ "${tid}" = "0x40" -o "${tid}" = "0x21" ]; then
				echo "*** Error: DK is missing.";
				exit 1;
			fi;
		fi;
	fi;

	if [ "${KEYFILE}" != "" ]; then
		if [ "${KEYVIADB}" != "" ]; then
			echo "*** Error: DB config does not accept keyfile.";
			exit 1;
		fi;
	elif [ "${KEYVIADB}" = "" ] && ([ "${Ctype}" = "PKC" ] || [ "${Ctype}" = "SBKPKC" ]); then
		if [ "${HASHFILE}" = "" ]; then
			if [ "${USEOPENSSL}" = "" ]; then
				KEYFILE="bootloader/rsa_priv.txt";
			else
				KEYFILE="bootloader/rsa_priv.pem";
			fi;
			KEYFILE=`readlink -f ${KEYFILE}`;
		fi;
	fi;
}

check_keks ()
{
	local name=$1;
	local value_old=$2;
	local value_new="$3";
	local tmp="";
	local slice_old="";
	local slice_new="";
	local -i i;
	local -i j;
	local -i len;

	if [ "${name}" = "KEK256" ]; then
		len=64;
	else
		len=32;
	fi;
	if [ -n "${value_new}" ]; then
		value_new=$(cat "${value_new}");
		tmp=$(echo "${value_new}" | cut -c1-2);
		if [ "${tmp}" = "0x" ] || [ "${tmp}" = "0X" ]; then
			value_new=$(echo "${value_new}" | cut -c3-);
		fi;
		i=${#value_new};
		if ((i < len)); then
			for ((j = 1; j <= len - i; j++)); do
				value_new="0"${value_new};
			done;
		fi;
	fi;
	tmp=$(echo ${value_old} | sed 's/^0*//');
	if [ -n "${tmp}" ] && [ -n "${value_new}" ]; then
		if [ ${force} -eq 1 ]; then
			for ((i = 1; i <= len; i = i + 8)); do
				j=i+7;
				slice_old=$(echo "${value_old}" | cut -c${i}-${j});
				slice_old="0x$(echo "obase=16; ibase=16; ${slice_old}" | bc)";
				slice_new=$(echo "${value_new}" | cut -c${i}-${j});
				slice_new="0x$(echo "obase=16; ibase=16; ${slice_new}" | bc)";
				tmp=$((${slice_old} & ${slice_new}));
				if ((${tmp} != ${slice_old})); then
					echo "${name}: you can't reset bits from 1 to 0. You're changing 0x${value_old} to 0x${value_new}.";
					return 1;
				fi;
			done;
		else
			echo "The ${name} has been burned. You can't burn it again.";
			echo "Add --force if you really want to do it.";
			return 1;
		fi;
	fi;

	return 0;
}

check_fuse_values ()
{
	local __auth=$1;
	local tmp="";
	local odm_lock_and="";
	local odm_old_name="";
	local odm_new_name="";
	local num="";
	declare -i mask=0;
	for f in ${FUSES_KEYS[@]}; do
		tmp=$(cat ${FUSE_READ_OUTPUT} |grep "${f}:"|cut -d" " -f2);
		eval "__fuse_check_${f}=${tmp}";
	done;
	for f in ${FUSES_MANUFACTURING[@]}; do
		tmp=$(cat ${FUSE_READ_OUTPUT} |grep "${f}:"|cut -d" " -f2);
		eval "__fuse_check_${f}=0x$(echo "obase=16; ibase=16; ${tmp}" | bc)";
	done;
	for f in ${FUSES_ODM_RESERVED[@]}; do
		tmp=$(cat ${FUSE_READ_OUTPUT} |grep "${f}:"|cut -d" " -f2);
		eval "__fuse_check_${f}=0x$(echo "obase=16; ibase=16; ${tmp}" | bc)";
	done;
	if [ "${tid}" = "0x19" ]; then
		for f in ${FUSES_T19X_EXTRA_ODM[@]}; do
			tmp=$(cat ${FUSE_READ_OUTPUT} |grep "${f}:"|cut -d" " -f2);
			eval "__fuse_check_${f}=0x$(echo "obase=16; ibase=16; ${tmp}" | bc)";
		done;
	fi;

	if ((${__fuse_check_SecurityMode} > 0)); then
		if [ -n "${KEK0FILE}" ] || [ -n "${KEK1FILE}" ] || [ -n "${KEK2FILE}" ] ||
		   [ -n "${KEK256FILE}" ] || [ -n "${BootDevCfg}" ] || [ ${jtag_disable} -eq 1 ] ||
		   [ -n "${set_productionmode}" ] || [ -n "${sw_reserved}" ] ||
		   [ -n "${arm_debug_auth}" ] || [ -n "${odm_id}" ]; then
			echo "Production mode is set, you can't burn any manufacturing fuses now.";
			return 1;
		fi;
		if [ "${__auth}" = "NS" ]; then
			if [ -n "${KEYFILE}" ] || [ -n "${SBKFILE}" ]; then
				echo "Production mode is set, you can't burn PKC or SBK now.";
				return 1;
			fi;
		fi;
	fi;
	# We don't need to check PKC and SBK here.
	# If PKC is burned, then the KEYFILE users provide is for signing the images
	# If SBK is burned, then the SBKFILE users provide is for encrypting the images
	# If both PKC and SBK are not burned, then the KEYFILE and SBKFILE users provide are for burning
	check_keks "KEK0" "${__fuse_check_Kek0}" "${KEK0FILE}" || { return 1; };
	check_keks "KEK1" "${__fuse_check_Kek1}" "${KEK1FILE}" || { return 1; };
	check_keks "KEK2" "${__fuse_check_Kek2}" "${KEK2FILE}" || { return 1; };
	check_keks "KEK256" "${__fuse_check_Kek256}" "${KEK256FILE}" || { return 1; };

	if [ ${jtag_disable} -eq 1 ] && ((${__fuse_check_JtagDisable} > 0)); then
		echo "The jtag-disable fuse bit has been burned. You can't burn it again.";
		return 1;
	fi;
	if [ -n "${sw_reserved}" ]; then
		tmp=$((${__fuse_check_SwReserved} & ${sw_reserved}));
		if ((${tmp} != ${__fuse_check_SwReserved})); then
			echo "SwReserved: you can't reset bits from 1 to 0. You're changing ${__fuse_check_SwReserved} to ${sw_reserved}.";
			return 1;
		fi;
	fi;
	if [ -n "${arm_debug_auth}" ]; then
		tmp=$((${__fuse_check_DebugAuthentication} & ${arm_debug_auth}));
		if ((${tmp} != ${__fuse_check_DebugAuthentication})); then
			echo -n "DebugAuthentication: you can't reset bits from 1 to 0. ";
			echo "You're changing ${__fuse_check_DebugAuthentication} to ${arm_debug_auth}.";
			return 1;
		fi;
	fi;
	if [ -n "${odm_id}" ]; then
		tmp=$((${__fuse_check_OdmId} & ${odm_id}));
		if ((${tmp} != ${__fuse_check_OdmId})); then
			echo "OdmId: you can't reset bits from 1 to 0. You're changing ${__fuse_check_OdmId} to ${odm_id}.";
			return 1;
		fi;
	fi;

	if [ -n "${odm_lock}" ]; then
		tmp=$((${__fuse_check_OdmLock} & ${odm_lock}));
		if ((${tmp} != ${__fuse_check_OdmLock})); then
			echo "OdmLock: you can't reset bits from 1 to 0. You're changing ${__fuse_check_OdmLock} to ${odm_lock}.";
			return 1;
		fi;
	fi;
	local odm_reserved_fuses=("${FUSES_ODM_RESERVED[@]}");
	for f in ${odm_reserved_fuses[@]}; do
		if [[ "${f}" = ReservedOdm* ]]; then
			num=${f##ReservedOdm};
			odm_new_name="odm_reserved${num}";
			if [ -n "${!odm_new_name}" ]; then
				if ((${num} < 4)); then
					let "mask = 1 << ${num}";
					odm_lock_and=$((${__fuse_check_OdmLock} & ${mask}));
					if ((${odm_lock_and} > 0)); then
						echo "OdmLock[${num}] has been set. You can't burn ${odm_new_name} again.";
						return 1;
					fi;
				fi;

				odm_old_name="__fuse_check_${f}";
				tmp=$((${!odm_old_name} & ${!odm_new_name}));
				if ((${tmp} != ${!odm_old_name})); then
					echo "${odm_new_name}: you can't reset bits from 1 to 0. You're changing ${!odm_old_name} to ${!odm_new_name}.";
					return 1;
				fi;
			fi;
		fi;
	done;

	return 0;
}

generate_key_hash ()
{
	local __hash="${1}";
	local __rsa_key_len="${2}";
	local chip="${3}";
	local ktype="${4}";
	local keyfile="${5}";
	local tmpdir;
	local rsa_modulus;
	local rsa_key_len;
	local buf_len;
	local result;
	local tegraopenssl="${BL_DIR}"/tegraopenssl;

	result="";
	if [ "${ktype}" = "rsa" ]; then
		tmpdir="$(mktemp -d generate-key-hash-tmpdir-XXXX)";
		pushd "${tmpdir}" >& /dev/null;
		openssl rsa -in "${keyfile}" -noout -modulus -out key.mod;
		cat key.mod | sed s/^Modulus=// > key.mod.tmp;
		rsa_modulus="$(cat key.mod.tmp)";
		xxd -r -p key.mod.tmp key.mod.tmp.bin;

		case ${#rsa_modulus} in
		$(( RSA_2K_SIZE * 2 )))
			# 2K RSA key
			objcopy -I binary --reverse-byte=${RSA_2K_SIZE} key.mod.tmp.bin key.mod.bin;
			rsa_key_len="${RSA_2K_SIZE}";
		;;
		$(( RSA_3K_SIZE * 2 )))
			if [ "${chip}" != "0x19" ]; then
				echo
				echo "Error: 3K RSA is only supported on Tegra194 platforms.";
				popd >& /dev/null;
				rm -rf "${tmpdir}";
				return 1;
			fi;
			${tegraopenssl} --isPkcKey "${keyfile}" key.hash mont.bin >& /dev/null;
			objcopy -I binary --reverse-byte=${RSA_3K_SIZE} key.mod.tmp.bin key.mod.bin;
			rsa_key_len="${RSA_3K_SIZE}";
		;;
		*)
			echo
			echo "Error: Unknown size RSA key: ${#rsa_modulus}.";
			popd >& /dev/null;
			rm -rf "${tmpdir}";
			return 1;
		;;
		esac;

		case "${chip}" in
		0x21) buf_len=256; ;;
		0x18) buf_len=896; ;;
		0x19) buf_len=1216; ;;
		*)
			echo
			echo "Error: Unsupported chip: ${chip}."
			return 1;
		;;
		esac;

		dd if=/dev/zero of=buffer.bin bs=1 count="${buf_len}" >& /dev/null;
		dd if=key.mod.bin conv=notrunc of=buffer.bin bs=1 >& /dev/null;
		if [ "${chip}" = "0x21" ]; then
			# Tegra210 uses little-endian so need an extra conversion here
			openssl dgst -sha256 -binary -out key.sha buffer.bin;
			objcopy -I binary --reverse-bytes=32 key.sha key.res;
			result=$(xxd -p -c 64 key.res);
		else
			if [ "${rsa_key_len}" = "${RSA_3K_SIZE}" ] && [ "${chip}" = "0x19" ]; then
				dd if=mont.bin conv=notrunc of=buffer.bin bs=1 seek=${RSA_3K_SIZE} >& /dev/null;
			fi;
			result=$(openssl dgst -sha256 -hex buffer.bin|cut -d" " -f2);
		fi;
		popd >& /dev/null;
		rm -rf "${tmpdir}";
	else
		# TODO: Add ECC and Ed25519 support here
		echo
		echo "Error: Unsupported key type: ${ktype}.";
		return 1;
	fi;

	eval ${__hash}="0x${result}";
	if [ "${ktype}" = "rsa" ]; then
		eval ${__rsa_key_len}="${rsa_key_len}";
	fi;
	return 0;
}

check_required_tools ()
{
	local tools=("bc" "xxd" "objcopy" "openssl");
	for t in ${tools[@]}; do
		if ! command -v "${t}" >& /dev/null; then
			echo "Command \"${t}\" cannot be found. Install it before continue.";
			return 1;
		fi;
	done;
	return 0;
}

check_required_tools || { exit 1; };

cd "$(dirname $0)";
source ./odmfuse.func;
LDK_DIR="$(pwd)";
LDK_DIR=`readlink -f "${LDK_DIR}"`;
BL_DIR="${LDK_DIR}/bootloader";
RSA_2K_SIZE=256
RSA_3K_SIZE=384

# For T210 platforms
#   The variable Ctype saves the value of the "-c" option
# For T186 and T194 platforms
#   The variable Ctype saves the value of the "-c" option originally.
#   We use it to save the value of option "--auth" now because the option "-c" has been obsoleted.
#   For offline mode, $Ctype is mandatory because we have no way to know the authentication status of the board.
#   For online and hybrid modes, $Ctype is optional because we always get the authentication status
#   of the board and save it to $bootauth. If users set $Ctype, we'll compare it with $bootauth and
#   fail the execution if they don't match.
Ctype="";
__ctype="";
noburn=0;
testmode=0;
force=0;
jtag_disable=0;
running_mode="";
arm_debug_auth="";
odm_id="";
while getopts "hc:d:i:jk:l:o:pr:s:D:H:S:X:-:" OPTION
do
	case $OPTION in
	h) usage; ;;
	c) __ctype=${OPTARG}; ;;
	d) BootDevCfg="${OPTARG}"; ;;
	i) tid="${OPTARG}"; pkcopt+="-i ${tid} "; ;;
	j)
		echo;
		echo "The option -j is obsolete now. Jtag by default is enabled.";
		echo "Please use \"--disable-jtag\" option if you want to burn the jtag-disable fuse.";
		echo "Jtag can't be re-enabled once the jtag-disable fuse bit is burned.";
		echo;
	;;
	k) KEYFILE="${OPTARG}"; ;;
	l) odm_lock="${OPTARG}"; ;;
	o) odm_reserved="${OPTARG}"; ;;
	p) set_productionmode="yes"; ;;
	r) sw_reserved="${OPTARG}"; ;;
	s) sku=${OPTARG}; pkcopt+="-s ${sku} "; ;;
	D) DKFILE="${OPTARG}"; ;;
	H) HASHFILE=${OPTARG}; ;;
	S) SBKFILE="${OPTARG}"; ;;
	X) XFILE="${OPTARG}"; ;;
	-) case ${OPTARG} in
	   noburn) noburn=1; ;;
	   test) testmode=1; ;;
	   force) force=1; ;;
	   disable-jtag) jtag_disable=1; ;;
	   auth) Ctype="${!OPTIND}"; OPTIND=$((OPTIND+1));;
	   KEK0) KEK0FILE="${!OPTIND}"; OPTIND=$((OPTIND+1));;
	   KEK1) KEK1FILE="${!OPTIND}"; OPTIND=$((OPTIND+1));;
	   KEK2) KEK2FILE="${!OPTIND}"; OPTIND=$((OPTIND+1));;
	   KEK256) KEK256FILE="${!OPTIND}"; OPTIND=$((OPTIND+1));;
	   odm_reserved0) odm_reserved0="${!OPTIND}"; OPTIND=$((OPTIND+1));;
	   odm_reserved1) odm_reserved1="${!OPTIND}"; OPTIND=$((OPTIND+1));;
	   odm_reserved2) odm_reserved2="${!OPTIND}"; OPTIND=$((OPTIND+1));;
	   odm_reserved3) odm_reserved3="${!OPTIND}"; OPTIND=$((OPTIND+1));;
	   odm_reserved4) odm_reserved4="${!OPTIND}"; OPTIND=$((OPTIND+1));;
	   odm_reserved5) odm_reserved5="${!OPTIND}"; OPTIND=$((OPTIND+1));;
	   odm_reserved6) odm_reserved6="${!OPTIND}"; OPTIND=$((OPTIND+1));;
	   odm_reserved7) odm_reserved7="${!OPTIND}"; OPTIND=$((OPTIND+1));;
	   debug_authentication) arm_debug_auth="${!OPTIND}"; OPTIND=$((OPTIND+1));;
	   odm_id) odm_id="${!OPTIND}"; OPTIND=$((OPTIND+1));;
	   *) usage; ;;
	   esac;;
	*) usage; ;;
	esac
done

sanitize_inputs;

shift $((OPTIND - 1));
if [ -f "${PWD}/pkc/keyviadb" ]; then
	KEYVIADB="yes";
	if [ $# -lt 2 ]; then
		usage;
	fi;
	master=$1;
	validateMaster ${master};
	sn=$2;

	if [ "${SUDO_USER}" != "" ]; then
		sudocmd="sudo -u ${SUDO_USER}";
	fi;

	if [ $# -gt 2 ]; then
		if [ "$3" == "all" ]; then
			colst="SN,ID,SKU,TYPE,FUSECNT, SIGNCNT, CTIME, ATIME";
			pkcmd="bash -l -c 'mkpkc getcol ${sn} \"${colst}\"'";
		else
			pkcmd="bash -l -c 'mkpkc getcol ${sn} $3'";
		fi;
		${sudocmd} ssh ${master} "${pkcmd}";
		exit $?;
	fi;
else
	if [ "${tid}" = "0x18" -o "${tid}" = "0x19" ]; then
		if [ $# -ne 1 ]; then
			usage;
		fi;
		if [ $# -eq 1 ]; then
			nargs=$#;
			ext_target_board=${!nargs};
			if [ ! -r ${ext_target_board}.conf ]; then
				echo -n "Error: Invalid target board - ";
				echo "${ext_target_board}";
				exit 1;
			fi;

			# set up environments:
			source ${ext_target_board}.conf
			TARGET_DIR="${BL_DIR}/${target_board}";
			KERNEL_DIR="${LDK_DIR}/kernel";
			DTB_DIR="${KERNEL_DIR}/dtb";
			odmfuse_init "${tid}" "${usb_instance}" "${CHIPMAJOR}" "${BL_DIR}" "${TARGET_DIR}" "${LDK_DIR}" "${SBKFILE}" "${KEYFILE}";

			if [ "${running_mode}" != "${MODE_OFFLINE}" ]; then
				# The fuse read function does 3 things:
				# (1) Call get_fuse_level to get the "bootauth" of the board
				# (2) Dump the eeprom of the board so that we can make fuse args
				# (3) Send nvtboot binaries to the board then read fuse values
				read_fuse_values bootauth CMDARGS;
				if [ $? -ne 0 ]; then
					echo "Error: read fuse from board failed.";
					exit 1;
				fi;
				if [ "${Ctype}" != "" ] && [ "${Ctype}" != "${bootauth}" ]; then
					echo "Error: wrong \"--auth\" option is set. The board's authentication type is: ${bootauth}.";
					exit 1;
				fi;

				pushd "${BL_DIR}" >& /dev/null;
				check_fuse_values "${bootauth}";
				if [ $? -ne 0 ]; then
					echo "Error: check fuse values failed.";
					exit 1;
				fi;
				popd >& /dev/null;
			else
				fuselevel="fuselevel_production";
				hwchipid="${tid}";
				bootauth="None";
				hwchiprev="0";

				declare -F -f process_fuse_level > /dev/null 2>&1;
				if [ $? -eq 0 ]; then
					process_fuse_level "${fuselevel}";
				fi;

				bd_ver="${FAB}";
				bd_id="${BOARDID}";
				bd_sku="${BOARDSKU}";
				bd_rev="${BOARDREV}";
				# process the board version and update the data accordingly
				declare -F -f process_board_version > /dev/null 2>&1;
				if [ "${CHIPREV}" != "" ]; then
					hwchiprev="${CHIPREV}";
				fi;
				process_board_version "${bd_id}" "${bd_ver}" "${bd_sku}" "${bd_rev}" "${hwchiprev}";
				prepare_cmdargs "${Ctype}" CMDARGS;
			fi;
		fi;
	else
		if [ $# -ne 0 ]; then
			usage;
		fi;
	fi;
	export PATH="${PWD}/pkc:${PATH}";
	if [ ! -f "${PWD}/pkc/sha2_key" ]; then
		USEOPENSSL="yes";
	fi;
fi;

if [ "${XFILE}" != "" ]; then
	if [ "${Ctype}" != "" -o \
	     "${BootDevCfg}" != "" -o \
	     ${jtag_disable} -eq 1 -o \
	     "${KEYFILE=}" != "" -o \
	     "${odm_lock}" != "" -o \
	     "${odm_reserved}" != "" -o \
	     "${set_productionmode}" != "" -o \
	     "${sw_reserved}" != "" -o \
	     "${DKFILE}" != "" -o \
	     "${SBKFILE}" != "" ]; then
		usage;
	fi;
	if [ ! -f "${XFILE}" ]; then
		echo "*** Error: ${XFILE} doesn't exits.";
		exit 1;
	fi;
	XFILE=`readlink -f "${XFILE}"`;
	pushd bootloader > /dev/null 2>&1;

	cp -f "${XFILE}" . > /dev/null 2>&1;
	fusecfg=`basename "${XFILE}"`;
	echo "*** Start fusing from fuse configuration ... ";
	if [ "${tid}" = "0x40" ]; then
		cp -f ardbeg/fastboot.bin .;
		fcmd="./nvflash --writefuse ${fusecfg} --bl fastboot.bin --go";
	elif [ "${tid}" = "0x21" ]; then
		fcmd="./tegraflash.py --chip ${tid} --applet nvtboot_recovery.bin ";
		fcmd+="--cmd \"blowfuses ${fusecfg};\"";
	elif [ "${tid}" = "0x18" ]; then
		fcmd="./tegraflash.py ${CMDARGS} ";
		fcmd+="--cmd \"burnfuses ${fusecfg}\"";
	elif [ "${tid}" = "0x19" ]; then
		fcmd="./tegraflash.py ${CMDARGS} ";
		fcmd+="--cmd \"burnfuses ${fusecfg}\"";
	fi;
	if [ $noburn -eq 1 ]; then
		factory_overlay_gen "${fcmd}" "${fusecfg}";
		exit $?;
	fi;
	echo "${fcmd}";
	eval ${fcmd};
	if [ $? -ne 0 ]; then
		echo "failed.";
		exit 1;
	fi;
	popd > /dev/null 2>&1;
	exit 0;
fi;

if [ "${KEYFILE}" != "" ]; then
	pkcopt+="-k ${KEYFILE} -f PKC ";
fi;
pushd bootloader > /dev/null 2>&1;
if [ "${KEYVIADB}" != "" ]; then
	echo -n "*** Requesting HASH from DB server ... ";
	hcmd="${sudocmd} ssh ${master}";
	hash=`${hcmd} "bash -l -c 'mkpkc ${pkcopt} genkey ${sn}'"`;
	if [[ "${hash}" =~ ^Error:[^@]{1,}$ ]]; then
		hash=`${hcmd} "bash -l -c 'mkpkc ${pkcopt} getcol ${sn} hash'"`;
		if [[ "${hash}" =~ ^Error:[^@]{1,}$ ]]; then
			echo "${hash}";
			echo "Error: failed to get key hash";
			exit 1;
		fi;
		hash="HASH: ${hash}";
	fi;
	echo "done";
else
	use_hash=0;
	if [ -f "${KEYFILE}" ]; then
		echo -n "*** Calculating HASH from keyfile ${KEYFILE} ... ";
	elif [ -f "${HASHFILE}" ]; then
		echo -n "*** Using HASH from ${HASHFILE} ... ";
		use_hash=1;
	fi;
	hash="";
	pkc_size="";
	if [ $use_hash -ne 1 ];  then
		if [ -f "${KEYFILE}" ]; then
			generate_key_hash hash pkc_size "${tid}" rsa "${KEYFILE}" || { exit 1; };
			echo "done";
			echo "PKC HASH: ${hash}";
		fi;
	else
		hash=`cat "${HASHFILE}"`;
		chkhash "hash" 32;
		echo "done";
	fi;
fi;

if [ "${KEK0FILE}" != "" ]; then
	kek0=`cat ${KEK0FILE}`;
	chkhash "kek0" 16;
fi;

if [ "${KEK1FILE}" != "" ]; then
	kek1=`cat ${KEK1FILE}`;
	chkhash "kek1" 16;
fi;

if [ "${KEK2FILE}" != "" ]; then
	kek2=`cat ${KEK2FILE}`;
	chkhash "kek2" 16;
fi;

if [ "${KEK256FILE}" != "" ]; then
	kek256=`cat ${KEK256FILE}`;
	chkhash "kek256" 32;
fi;

if [ "${DKFILE}" != "" ]; then
	DK=`cat "${DKFILE}"`;
	chkhash "DK" 4;
fi;

if [ "${SBKFILE}" != "" ]; then
	SBK="$(cat "${SBKFILE}" | sed -e s/^[[:space:]]*// -e s/[[:space:]]0x//g -e s/[[:space:]]*//g)";
	chkhash "SBK" 16;
fi;

echo -n "*** Generating fuse configuration ... ";
pkc_disable=0;
pkc_hash="";
# Handle chip 0x21 and 0x40 first
if [ "${Ctype}" = "NS" ]; then
	if [ "${tid}" = "0x21" ]; then
		fusecfg="odmfuse_pkc.xml";
	elif [ "${tid}" = "0x40" ]; then
		fusecfg="odmfuse_ns.cfg";
	fi;
	pkc_disable=1;
	pkc_hash="";
else
	if [ "${tid}" = "0x40" ]; then
		fusecfg="odmfuse_pkc.cfg";
	elif [ "${tid}" = "0x21" ]; then
		fusecfg="odmfuse_pkc.xml";
	fi;
	pkc_disable=0;
	pkc_hash="${hash}";
fi;

if [ "${tid}" = "0x21" ] && [ "${SBK}" != "" ] && [ ${pkc_disable} -eq 1 ]; then
	echo
	echo "SBK can't be used individually on Jetson TX1 and Jetson nano.";
	echo "SBK must be used along with the PKC key.";
	exit 0;
fi;

# For T186 and T194, bootauth is set for ONLINE and HYBRID mode
# Ctype is set for OFFLINE mode. And only a NS board can burn
# the PublicKeyHash fuse
if [ "${tid}" = "0x18" ] || [ "${tid}" = "0x19" ]; then
	fusecfg="odmfuse_pkc.xml";
	# Reset pkc_hash because it might be set when handling T210 and T40 above
	pkc_hash="";
	if [ "${Ctype}" = "NS" ] || [ "${bootauth}" = "NS" ]; then
		if [ "${KEYFILE}" != "" ]; then
			pkc_hash="${hash}";
		fi;
	fi;
fi;

rm -f ${fusecfg};
if [ "${tid}" = "0x40" ]; then
	if [ "${BootDevCfg}" != "" ]; then
		echo -n "<fuse:sec_boot_dev_cfg;" >> ${fusecfg};
		echo "value:${BootDevCfg}>" >> ${fusecfg};
		fusecnt=$((fusecnt + 1));
	fi;
	if [ ${jtag_disable} -eq 1 ]; then
		echo "<fuse:jtag_disable;value:0x1>" >> ${fusecfg};
		fusecnt=$((fusecnt + 1));
	fi;
	if [ "${odm_lock}" != "" ]; then
		echo "<fuse:odm_lock;value:${odm_lock}>" >> ${fusecfg};
		fusecnt=$((fusecnt + 1));
	fi;
	if [ "${odm_reserved}" != "" ]; then
		echo "<fuse:odm_reserved;value:${odm_reserved}>" >> ${fusecfg};
		fusecnt=$((fusecnt + 1));
	fi;
	if [ "${sw_reserved}" != "" ]; then
		echo "<fuse:sw_reserved;value:${sw_reserved}>" >> ${fusecfg};
		fusecnt=$((fusecnt + 1));
	fi;
	if [ "${DK}" != "" ]; then
		echo "<fuse:device_key;value:${DK}>"  >> ${fusecfg};
		fusecnt=$((fusecnt + 1));
	fi;
	if [ "${SBK}" != "" ]; then
		echo "<fuse:secure_boot_key;value:${SBK}>"  >> ${fusecfg};
		fusecnt=$((fusecnt + 1));
	fi;
	if [ ${pkc_disable} -eq 1 ]; then
		echo "<fuse:pkc_disable;value:0x1>"  >> ${fusecfg};
		fusecnt=$((fusecnt + 1));
	fi;
	if [ "${pkc_hash}" != "" ]; then
		echo "<fuse:public_key_hash;value:${pkc_hash}>"  >> ${fusecfg};
		fusecnt=$((fusecnt + 1));
	fi;
	if [ "${set_productionmode}" != "" ]; then
		echo "<fuse:odm_production_mode;value:0x1>"  >> ${fusecfg};
		fusecnt=$((fusecnt + 1));
	fi;
elif [ "${tid}" = "0x21" -o "${tid}" = "0x18" -o "${tid}" = "0x19" ]; then
	magicid="0x46555345";
	if [ "${tid}" = "0x18" -o "${tid}" = "0x19" ]; then
		magicid="0x45535546";	# BigEndian format
	fi;
	echo -n "<genericfuse " >> ${fusecfg};
	echo    "MagicId=\"${magicid}\" version=\"1.0.0\">" >> ${fusecfg};
	if [ "${BootDevCfg}" != "" ]; then
		echo -n "<fuse name=\"SecBootDeviceSelect\" " >> ${fusecfg};
		echo    "size=\"4\" value=\"${BootDevCfg}\" />" >> ${fusecfg};
		fusecnt=$((fusecnt + 1));
	fi;
	if [ ${jtag_disable} -eq 1 ]; then
		echo -n "<fuse name=\"JtagDisable\" " >> ${fusecfg};
		echo    "size=\"4\" value=\"0x1\" />" >> ${fusecfg};
		fusecnt=$((fusecnt + 1));
	fi;
	if [ "${odm_lock}" != "" ]; then
		echo -n "<fuse name=\"OdmLock\" " >> ${fusecfg};
		echo    "size=\"4\" value=\"${odm_lock}\" />" >> ${fusecfg};
		fusecnt=$((fusecnt + 1));
	fi;
	if [ "${odm_reserved}" != "" ]; then
		echo -n "<fuse name=\"ReservedOdm\" " >> ${fusecfg};
		echo "size=\"32\" value=\"${odm_reserved}\" />" >> ${fusecfg};
		fusecnt=$((fusecnt + 1));
	fi;
	if [ "${odm_reserved0}" != "" ]; then
		echo -n "<fuse name=\"ReservedOdm0\" " >> ${fusecfg};
		echo "size=\"4\" value=\"${odm_reserved0}\" />" >> ${fusecfg};
		fusecnt=$((fusecnt + 1));
	fi;
	if [ "${odm_reserved1}" != "" ]; then
		echo -n "<fuse name=\"ReservedOdm1\" " >> ${fusecfg};
		echo "size=\"4\" value=\"${odm_reserved1}\" />" >> ${fusecfg};
		fusecnt=$((fusecnt + 1));
	fi;
	if [ "${odm_reserved2}" != "" ]; then
		echo -n "<fuse name=\"ReservedOdm2\" " >> ${fusecfg};
		echo "size=\"4\" value=\"${odm_reserved2}\" />" >> ${fusecfg};
		fusecnt=$((fusecnt + 1));
	fi;
	if [ "${odm_reserved3}" != "" ]; then
		echo -n "<fuse name=\"ReservedOdm3\" " >> ${fusecfg};
		echo "size=\"4\" value=\"${odm_reserved3}\" />" >> ${fusecfg};
		fusecnt=$((fusecnt + 1));
	fi;
	if [ "${odm_reserved4}" != "" ]; then
		echo -n "<fuse name=\"ReservedOdm4\" " >> ${fusecfg};
		echo "size=\"4\" value=\"${odm_reserved4}\" />" >> ${fusecfg};
		fusecnt=$((fusecnt + 1));
	fi;
	if [ "${odm_reserved5}" != "" ]; then
		echo -n "<fuse name=\"ReservedOdm5\" " >> ${fusecfg};
		echo "size=\"4\" value=\"${odm_reserved5}\" />" >> ${fusecfg};
		fusecnt=$((fusecnt + 1));
	fi;
	if [ "${odm_reserved6}" != "" ]; then
		echo -n "<fuse name=\"ReservedOdm6\" " >> ${fusecfg};
		echo "size=\"4\" value=\"${odm_reserved6}\" />" >> ${fusecfg};
		fusecnt=$((fusecnt + 1));
	fi;
	if [ "${odm_reserved7}" != "" ]; then
		echo -n "<fuse name=\"ReservedOdm7\" " >> ${fusecfg};
		echo "size=\"4\" value=\"${odm_reserved7}\" />" >> ${fusecfg};
		fusecnt=$((fusecnt + 1));
	fi;
	if [ "${tid}" = "0x18" -o "${tid}" = "0x19" ]; then
		if [ "${arm_debug_auth}" != "" ]; then
			echo -n "<fuse name=\"DebugAuthentication\" " >> ${fusecfg};
			echo "size=\"4\" value=\"${arm_debug_auth}\" />" >> ${fusecfg};
			fusecnt=$((fusecnt + 1));
		fi;

		if [ "${odm_id}" != "" ]; then
			echo -n "<fuse name=\"OdmId\" " >> ${fusecfg};
			echo "size=\"8\" value=\"${odm_id}\" />" >> ${fusecfg};
			fusecnt=$((fusecnt + 1));
		fi;
	fi;
	if [ "${sw_reserved}" != "" ]; then
		echo -n "<fuse name=\"SwReserved\" " >> ${fusecfg};
		echo "size=\"4\" value=\"${sw_reserved}\" />" >> ${fusecfg};
		fusecnt=$((fusecnt + 1));
	fi;
	if [ "${DK}" != "" ]; then
		echo -n "<fuse name=\"DeviceKey\" " >> ${fusecfg};
		echo    "size=\"8\" value=\"${DK}\" />"  >> ${fusecfg};
		fusecnt=$((fusecnt + 1));
	fi;
	if [ "${SBK}" != "" ]; then
		if [ "${tid}" = "0x18" ] || [ "${tid}" = "0x19" ]; then
			if ([ "${bootauth}" = "NS" ] || [ "${Ctype}" = "NS" ]); then
				echo -n "<fuse name=\"SecureBootKey\" " >> ${fusecfg};
				echo    "size=\"16\" value=\"${SBK}\" />"  >> ${fusecfg};
				fusecnt=$((fusecnt + 1));
			fi;
		else
			echo -n "<fuse name=\"SecureBootKey\" " >> ${fusecfg};
			echo    "size=\"16\" value=\"${SBK}\" />"  >> ${fusecfg};
			fusecnt=$((fusecnt + 1));
		fi;
	fi;
	if [ "${kek0}" != "" ]; then
		echo -n "<fuse name=\"Kek0\" " >> ${fusecfg};
		echo    "size=\"16\" value=\"${kek0}\" />"  >> ${fusecfg};
		fusecnt=$((fusecnt + 1));
	fi;
	if [ "${kek1}" != "" ]; then
		echo -n "<fuse name=\"Kek1\" " >> ${fusecfg};
		echo    "size=\"16\" value=\"${kek1}\" />"  >> ${fusecfg};
		fusecnt=$((fusecnt + 1));
	fi;
	if [ "${kek2}" != "" ]; then
		echo -n "<fuse name=\"Kek2\" " >> ${fusecfg};
		echo    "size=\"16\" value=\"${kek2}\" />"  >> ${fusecfg};
		fusecnt=$((fusecnt + 1));
	fi;
	if [ "${kek256}" != "" ]; then
		echo -n "<fuse name=\"Kek256\" " >> ${fusecfg};
		echo    "size=\"32\" value=\"${kek256}\" />"  >> ${fusecfg};
		fusecnt=$((fusecnt + 1));
	fi;
	if [ "${tid}" = "0x21" ] && [ ${pkc_disable} -eq 1 ]; then
		echo "<fuse name=\"PkcDisable\" size=\"4\" value=\"0x1\" />" >> ${fusecfg};
		fusecnt=$((fusecnt + 1));
	fi;
	if [ "${pkc_hash}" != "" ]; then
		echo -n "<fuse name=\"PublicKeyHash\" size=\"32\" " >> ${fusecfg};
		echo    "value=\"${pkc_hash}\" />" >> ${fusecfg};
		fusecnt=$((fusecnt + 1));
		if [ "${tid}" = "0x18" ]; then
			echo -n "<fuse name=" >> ${fusecfg};
			echo -n "\"BootSecurityInfo\" " >> ${fusecfg};
			if [ "${SBK}" != "" ]; then
				bsi="0x6";
			else
				bsi="0x2";
			fi;
			echo    "size=\"4\" value=\"${bsi}\" />"  >> ${fusecfg};
			fusecnt=$((fusecnt + 1));
		elif [ "${tid}" = "0x19" ]; then
			echo -n "<fuse name=" >> ${fusecfg};
			echo -n "\"BootSecurityInfo\" " >> ${fusecfg};
			case ${pkc_size} in
			${RSA_2K_SIZE})
				if [ "${SBK}" != "" ]; then
					bsi="0x5";
				else
					bsi="0x1";
				fi;
			;;
			${RSA_3K_SIZE})
				if [ "${SBK}" != "" ]; then
					bsi="0x6";
				else
					bsi="0x2";
				fi;
			;;
			esac;
			echo "size=\"4\" value=\"${bsi}\" />"  >> ${fusecfg};
			fusecnt=$((fusecnt + 1));
		fi;
	fi;
	if [ "${set_productionmode}" != "" ]; then
		if [ "${tid}" = "0x21" -o "${tid}" = "0x18" -o \
			"${tid}" = "0x19" ]; then
			echo -n "<fuse name=\"SecurityMode\" " >> ${fusecfg};
			echo    "size=\"4\" value=\"0x1\" />"  >> ${fusecfg};
		else
			echo -n "<fuse name=\"ProductionMode\" " >> ${fusecfg};
			echo    "size=\"4\" value=\"0x1\" />"  >> ${fusecfg};
		fi;
		fusecnt=$((fusecnt + 1));
	fi;
	echo "</genericfuse>" >> ${fusecfg};
fi;
echo "done.";

if [ $fusecnt -eq 0 ]; then
	echo "*** No fuse bit specified. Terminating.";
	exit 0;
fi;

cp "${fusecfg}" "${fusecfg}.sav";

if [ "${tid}" = "0x40" ]; then
	cp -f ardbeg/fastboot.bin .;
	fcmd="./nvflash --writefuse ${fusecfg} --bl fastboot.bin --go";
elif [ "${tid}" = "0x21" ]; then
	if [ ${testmode} -eq 1 ]; then
		echo "Test mode: removing all lines with '<fuse name=' so no fuses will be burned.";
		sed -i '/<fuse name=/d' "${fusecfg}";
	fi;
	fcmd="./tegraflash.py --chip ${tid} --applet nvtboot_recovery.bin ";
	fcmd+="--cmd \"blowfuses ${fusecfg}; reboot recovery\"";
elif [ "${tid}" = "0x18" -o "${tid}" = "0x19" ]; then
	fcmd="./tegraflash.py ${CMDARGS} ";
	if [ "${RAMCODE}" != "" ]; then
		fcmd+=" --ramcode ${RAMCODE} ";
	fi;

	# For T186 and T194, tegraflash_internal.py always reboots to recovery
	# when fuse burning is done. So we don't add the "reboot recovery" in
	# the tegraflash command here.
	if [ ${testmode} -eq 1 ]; then
		echo "Test mode: using dummy so no fuses will be burned.";
		fcmd+="--cmd \"burnfuses dummy\"";
	else
		fcmd+="--cmd \"burnfuses ${fusecfg}\"";
	fi;
fi;

if [ $noburn -eq 1 ]; then
	factory_overlay_gen "${fcmd}" "${fusecfg}";
	exit $?;
fi;

echo "*** Start fusing ${sn} ... ";
echo "${fcmd}";
eval ${fcmd};
if [ $? -ne 0 ]; then
	echo "failed.";
	exit 1;
fi;
echo "*** The fuse configuration is saved in bootloader/${fusecfg}";
echo "*** The ODM fuse has been burned successfully.";
popd > /dev/null 2>&1;
echo "*** done.";
exit 0;
